#
# minimal implementation of the linear spin-wave algorithm (https://arxiv.org/abs/1402.6069) for comparison
# @author Tobias Weber <tweber@ill.fr>
# @date 24-oct-2024
# @license GPLv3, see 'LICENSE' file
#

import numpy as np
import numpy.linalg as la


verbose_print = False    # print intermediate results


#
# debug output
#
def print_infos(str):
	if verbose_print:
		print(str)


#
# skew-symmetric (cross-product) matrix
#
def skew(vec):
	return np.array([
		[      0.,   vec[2],  -vec[1] ],
		[ -vec[2],       0.,   vec[0] ],
		[  vec[1],  -vec[0],       0. ] ])


#
# calculate structure properties
#
def init(sites, couplings):
	# calculate spin rotations towards ferromagnetic order along [001]
	for site in sites:
		zdir = np.array([ 0., 0., 1. ])
		Sdir = np.array(site["Sdir"]) / la.norm(site["Sdir"])
		rotaxis = np.array([ 0., 1., 0. ])
		s = 0.

		if np.allclose(Sdir, zdir):
			# spin and z axis parallel
			c = 1.
		elif np.allclose(Sdir, -zdir):
			# spin and z axis anti-parallel
			c = -1.
		else:
			# sine and cosine of the angle between spin and z axis
			rotaxis = np.cross(Sdir, zdir)
			s = la.norm(rotaxis)
			c = np.dot(Sdir, zdir)
			rotaxis /= s

		# rotation via rodrigues' formula, see (Arens 2015), p. 718 and p. 816
		rot = (1. - c) * np.outer(rotaxis, rotaxis) + np.diag([ c, c, c ]) - skew(rotaxis)*s
		site["u"] = rot[0, :] + 1j * rot[1, :]
		site["v"] = rot[2, :]

		print_infos(np.dot(rot, Sdir))
		print_infos("\nrot = \n%s\nu = %s\nv = %s" % (rot, site["u"], site["v"]))


	# calculate real interaction matrices
	for coupling in couplings:
		J = coupling["J"]
		coupling["J_real"] = np.diag([ J, J, J ]) + skew(coupling["DMI"])
		print_infos("\nJ_real =\n%s" % coupling["J_real"])


#
# get the energies of the dispersion at the momentum transfer Qvec
#
def get_energies(Qvec, sites, couplings):
	print_infos("\n\nQ = %s" % Qvec)

	# fourier transform interaction matrices
	num_sites = len(sites)
	J_fourier = np.zeros((num_sites, num_sites, 3, 3), dtype = complex)
	J0_fourier = np.zeros((num_sites, num_sites, 3, 3), dtype = complex)

	for coupling in couplings:
		dist = np.array(coupling["dist"])
		J_real = coupling["J_real"]
		site1 = coupling["sites"][0]
		site2 = coupling["sites"][1]

		J_ft = J_real * np.exp(-1j * 2.*np.pi * np.dot(dist, Qvec))
		J_fourier[site1, site2] += J_ft
		J_fourier[site2, site1] += J_ft.transpose().conj()
		J0_fourier[site1, site2] += J_real
		J0_fourier[site2, site1] += J_real.transpose().conj()

	print_infos("\nJ_fourier =\n%s\n\nJ0_fourier =\n%s" % (J_fourier, J0_fourier))


	# hamiltonian
	H = np.zeros((2*num_sites, 2*num_sites), dtype = complex)

	for i in range(num_sites):
		S_i = sites[i]["S"]
		u_i = sites[i]["u"]
		v_i = sites[i]["v"]

		for j in range(num_sites):
			S_j = sites[j]["S"]
			u_j = sites[j]["u"]
			v_j = sites[j]["v"]
			S = 0.5 * np.sqrt(S_i * S_j)

			H[i, j] += S * np.dot(u_i, np.dot(J_fourier[i, j], u_j.conj()))
			H[i, i] -= S_j * np.dot(v_i, np.dot(J0_fourier[i, j], v_j))
			H[num_sites + i, num_sites + j] += \
				S * np.dot(u_i.conj(), np.dot(J_fourier[i, j], u_j))
			H[num_sites + i, num_sites + i] -= \
				S_j * np.dot(v_i, np.dot(J0_fourier[i, j], v_j))
			H[i, num_sites + j] += \
				S * np.dot(u_i, np.dot(J_fourier[i, j], u_j))
			H[num_sites + i, j] += \
				(S * np.dot(u_j, np.dot(J_fourier[j, i], u_i))).conj()

	print_infos("\nH =\n%s" % H)


	# trafo
	C = la.cholesky(H)
	signs = np.diag(np.concatenate((np.repeat(1, num_sites), np.repeat(-1, num_sites))))
	H_trafo = np.dot(C.transpose().conj(), np.dot(signs, C))
	print_infos("\nC =\n%s\n\nH_trafo =\n%s" % (C, H_trafo))


	# the eigenvalues of H give the energies
	Es = np.real(la.eigvals(H_trafo))
	print_infos("\nEs = %s" % Es)

	return Es


# test
if __name__ == "__main__":
	is_ferromagnetic = True  # choose ferromagnetic or antiferromagnetic 1d spin chain
	only_pos_E = True        # hide magnon annihilation?

	#
	# set up magnetic sites and couplings:
	#
	#   - magnetic sites:
	#     "S": spin magnitude
	#     "Sdir": spin direction
	#
	#   - magnetic couplings:
	#     "sites": indices of the sites to couple
	#     "J": (symmetric) exchange interaction
	#     "DMI": (antisymmetric) Dzyaloshinskii-Moryia interaction
	#     "dist": distance in rlu to the next unit cell for the coupling
	#
	if is_ferromagnetic:
		sites = [
			{ "S" : 1., "Sdir" : [ 0, 0, 1 ] },
		]

		couplings = [
			{ "sites" : [ 0, 0 ], "J" : -1., "DMI" : [ 0, 0, 0 ], "dist" : [ 1, 0, 0 ] },
		]
	else:  # antiferromagnetic
		sites = [
			{ "S" : 1., "Sdir" : [ 0, 0, 1 ] },
			{ "S" : 1., "Sdir" : [ 0, 0, -1 ] },
		]

		couplings = [
			{ "sites" : [ 0, 1 ], "J" : 1., "DMI" : [ 0, 0, 0 ], "dist" : [ 0, 0, 0 ] },
			{ "sites" : [ 1, 0 ], "J" : 1., "DMI" : [ 0, 0, 0 ], "dist" : [ 2, 0, 0 ] },
		]

	init(sites, couplings)


	# calculate a single point
	#get_energies([0.2, 0., 0.])
	#exit(0)

	# plot a dispersion branch
	import matplotlib.pyplot as plt

	hs = []
	Es = []
	for h in np.linspace(-1, 1, 1024):
		try:
			Qvec = np.array([ h, 0, 0 ])
			for E in get_energies(Qvec, sites, couplings):
				if only_pos_E and E < 0.:
					continue
				hs.append(h)
				Es.append(E)
		except la.LinAlgError:
			pass

	plt.plot()
	plt.xlabel("h (rlu)")
	plt.ylabel("E (meV)")
	plt.scatter(hs, Es, marker = '.')
	plt.show()
